; aid routines for field creation, retrieving, etc..
; written for DATAdesign engine v2.+
; (c)1992 by Joachim Van der Auwera for PROGS
        section text
        include win1_dd_src_engine_aid_mac
        include win1_dd_src_engine_struct_h
        include win1_inc_keys_err
even    equ     $fffffffe       ; value to AND to make things even
*
; find field in record with given length and start address
; find given fieldid, return startaddress or error code
; parameters :  d1      length
;               a1      address of record
;               d2      fieldid
; return        d0      error code
;               a1      address of start of field
; errors
; itnf  field not found in this record
findfld xsub
        regsave d1/d3
        moveq   #0,d0           ; assume everything is ok

        tst.l   d1              ; empty record ?
        beq.s   itnf            ; yep -> ..

floop   cmp.b   rf_id(a1),d2    ; is this the field we're looking for
        beq.s   fldend          ; fep -> found it

        move.l  rf_len(a1),d3   ; length of this field
        addi.l  #rf_fld,d3      ; and length of field header
        addq.l  #1,d3           ; make it larger
        andi.l  #even,d3        ; and even
        sub.l   d3,d1           ; remaining length of record
        ble.s   itnf            ; we checked all fields and couldn't find it
        adda.l  d3,a1           ; start of next field
        bra.s   floop

itnf    moveq   #err.itnf,d0    ; no we couldn't find it
fldend
        subend
*
; create a new field in the record buffer
; adjusts b_rlen and possibly b_mlen, b_mem
; fills in rf_id, rf_len, b_change
; parameters :  d1      length of field
;               d2      fieldid
;               a4      bufferid
; return        d0      error code
;               a1      start of new fieldheader
; errors
; imem  insufficient memory (old situation is preserved)
newfld  xsub
        regsave d3/d4/a0/a2/a3
        move.b  #1,b_change(a4) ; we are changing something to the record

        move.l  b_rlen(a4),d3   ; old record len
        add.l   d1,d3           ; new record len
        addi.l  #rf_fld,d3      ; which has to include fieldheader
        addi.l  #1,d3           ; record len always has to be larger
        andi.l  #even,d3        ; and even

        cmp.l   b_mlen(a4),d3   ; is our buffer large enough
        bgt.s   nfbmem          ; no -> need to allocate a new buffer

        ; fill in new record len
        move.l  b_rlen(a4),d4   ; old record len
        move.l  d3,b_rlen(a4)   ; new record len
        move.l  b_mem(a4),a1    ; start of record buffer
        adda.l  d4,a1           ; start of new field
        ; fill in field header
fillfld move.b  d2,rf_id(a1)    ; fieldid
        clr.b   rf_empty(a1)    ; clear UNUSED
        move.l  d1,rf_len(a1)   ; length of field

        moveq   #0,d0
        bra.s   nfend

nfbmem  ; allocate new buffer memory and copy old contents
        move.l  d3,d4           ; keep new record len
        addi.l  #511,d3         ; round buffer length up by 512 bytes
        andi.l  #$ffffffe0,d3
        exg     d1,d3           ; swap rec/fieldlen
        xjsr    getmem          ; reserve the mem for it
        exg     d1,d3           ; swap again
        bmi.s   nfend           ; someone has been memory gready again ->

        move.l  d3,b_mlen(a4)   ; new length of buffer memory
        move.l  b_mem(a4),a2    ; start of old buffer
        move.l  a2,a3           ; and keep it for a while
        move.l  a0,b_mem(a4)    ; fill in start of new buffer
        move.l  b_rlen(a4),d0   ; length of old record
        move.l  d4,b_rlen(a4)   ; new length
        move.l  a0,a1           ; new buffer
        adda.l  d0,a1           ; start of new field

        lsr.l   #1,d0           ; nr of words to copy
cpyrec  move.w  (a2)+,(a0)+     ; copy a word
        subq.l  #1,d0           ; another one ?
        bgt.s   cpyrec

        move.l  a3,a0           ; start of old buffer
        xjsr    relmem          ; release it to the system

        bra     fillfld         ; fill in field header

nfend
        subend
*
; increase (decrease) size of a field in the record buffer
; the allocated area is added at the end of the field or record for a new field
; the released area is deleted at the end of the field
; takes rules cf fieldheader at even address in account
; adjusts b_rlen, rf_len, b_change and possibly b_mlen, b_mem
; parameters :  d1      nr of bytes to allocate (or release if negative)
;               a1      start of field to increase or decrease
;               a4      bufferid
; return        d0      error code
;               a1      start of field
; errors
; imem  insufficient memory (old situation is preserved)
incrfld xsub
        regsave d1-d5/a0/a2/a3
s.incr  equ     0       .l ;nr of bytes to allocate
s.rlen  equ     4       .l ;new recordlen
s.faddr equ     8       .l ;address of field
s.len   equ     12
        getssp  s.len
        move.b  #1,b_change(a4) ; we are changing something to the record

        moveq   #0,d0           ; assume no problems
        move.l  a1,s.faddr(sp)  ; start address of field
        move.l  d1,s.incr(sp)   ; we want to do something
        beq     codend          ; ok, if you insist

        move.l  rf_len(a1),d3   ; length of the field we want to change
        andi.l  #1,d3           ; we only need bit 0
        eori.w  #1,d3           ; change sign
        add.l   b_rlen(a4),d3   ; old record len
        add.l   d1,d3           ; size change
        andi.l  #even,d3        ; the record len has to be even (d3=new record size)
        move.l  d3,s.rlen(sp)   ; fill in

        cmp.l   b_mlen(a4),d3   ; is our buffer large enough
        bgt.s   newmem          ; no -> need to allocate a new buffer

        move.l  rf_len(a1),d4   ; length of field
        addq.l  #1,d4           ; round up
        andi.l  #even,d4        ; make it even
        lea     rf_fld(a1,d4.l),a2      ; place where we have to start shifting

        move.l  rf_len(a1),d4   ; length of field
        eor.l   d1,d4           ; length increase eoren
        andi.l  #1,d4           ; we only need bit 0 to get what we have to add before shifting

        move.l  rf_len(a1),d5
        add.l   d1,d5           ; get new field length

        add.l   d4,d1
        andi.l  #even,d1        ; nr of bytes we have to skip when copying

        ; calculate the nr of words we have to shift
        move.l  b_mem(a4),d2    ; start of record buffer
        add.l   b_rlen(a4),d2   ; end of record
        sub.l   a2,d2           ; nr of bytes to copy
        lsr.l   #1,d2           ; nr of words to copy

        ; fill in new field and record len
        move.l  s.rlen(sp),b_rlen(a4)
        move.l  d5,rf_len(a1)

        tst.l   s.incr(sp)      ; do we have to increase the length ?
        bmi.s   decrfld         ; no ->..

        adda.l  d2,a2           ; start copying from the back
        adda.l  d2,a2

incfld  move.w  -(a2),0(a2,d1.l); copy a word
        subq.l  #1,d2           ; another one ?
        bgt.s   incfld
        bra.s   codend

decrfld move.w  (a2)+,-2(a2,d1.l);copy a word
        subq.l  #1,d2           ; another one ?
        bgt.s   decrfld
codend
        subend

newmem  ; allocate new buffer and fill it in (increasing buffer size)
        move.l  d3,d4           ; keep new record len
        addi.l  #511,d3         ; round buffer length up by 512 bytes
        andi.l  #$ffffffe0,d3
        exg     d1,d3           ; swap rec/fieldlen
        xjsr    getmem          ; reserve the mem for it
        exg     d1,d3           ; swap again
        bmi.s   codend          ; someone has been memory ready again ->

        move.l  d3,b_mlen(a4)   ; new length of buffer memory
        move.l  b_mem(a4),a2    ; start of old buffer
        move.l  a2,a3           ; and keep a copy
        move.l  a0,b_mem(a4)    ; fill in start of new buffer

        move.l  rf_len(a1),d1   ; old length of field
        addq.l  #1,d1
        andi.l  #even,d1        ; round up to even value
        addi.l  #rf_fld,d1      ; add fieldheader length
        suba.l  a2,a1           ; offset of field in record
        add.l   a1,d1           ; nr of bytes to copy before field
        adda.l  a0,a1           ; new start of field
        move.l  b_rlen(a4),d2   ; old length of record
        sub.l   d1,d2           ; nr of bytes after field to copy

        move.l  s.rlen(sp),d3   ; new length
        sub.l   d1,d3           ; minus beginning
        sub.l   d2,d3           ; and end, gives nr of bytes to skip
        lsr.l   #1,d1           ; nr of words to copy
        lsr.l   #1,d2           ; nr of words to copy

bgcpy   move.w  (a2)+,(a0)+     ; copy a word
        subq.l  #1,d1           ; another one ?
        bgt.s   bgcpy
        adda.l  d3,a0           ; skip reserved area
endcpy  move.w  (a2)+,(a0)+     ; copy a word
        subq.l  #1,d2           ; another one ?
        bgt.s   endcpy

        ; fill in new field and record len
        move.l  rf_len(a1),d5   ; old fieldlen
        add.l   s.incr(sp),d5   ; add increment
        move.l  s.rlen(sp),b_rlen(a4)
        move.l  d5,rf_len(a1)

        move.l  a3,a0           ; start of old buffer
        xjsr    relmem          ; release it to the system

        bra     codend          ; this is all we have to do
*
; delete field in record in buffer
; adjusts b_change, b_rlen if field existed
; parameters :  d2      fieldid
;               a4      bufferid
; return        d0      0
delfld  xsub
        regsave d1/d3/a1
        moveq   #0,d0           ; assume everything is ok

        move.l  b_rlen(a4),d1   ; record len
        beq.s   delend          ; zero -> nothing to do
        move.l  b_mem(a4),a1    ; area with record

dellp   move.l  rf_len(a1),d3   ; length of this field
        addi.l  #rf_fld,d3      ; and length of field header
        addq.l  #1,d3           ; make it larger
        andi.l  #even,d3        ; and even
        sub.l   d3,d1           ; remaining length of record (nr of bytes left in record)
        ble.s   delend          ; we checked all fields and couldn't find it -> nothing to delete
        cmp.b   rf_id(a1),d2    ; is this the field we're looking for
        beq.s   dodel           ; yep -> found it
        adda.l  d3,a1           ; start of next field
        bra.s   dellp
dodel   move.b  0(a1,d3.l),(a1)+; erase field
        subq.l  #1,d1           ; another byte left in record ?
        bgt.s   dodel           ; -> copy another
        sub.l   d3,b_rlen(a4)   ; new length of record
        move.b  #1,b_change(a4) ; the record has changed
delend  subend
;
        end
